#!/usr/env/bin python

import enum
import logging
import os
import typing

from urllib.parse import urlparse

NOOP: str = 'noop'
DNS: str = '8.8.8.8'
ENCODING: str = 'utf-8'
DEBUG: bool = False if os.environ.get('DEBUG', 'true') in ['f', 'false', 'no'] else True
DELAY: int = .1

REDIS_URL: str = os.environ.get('REDIS_URL', 'http://localhost:6379/4')
_redis_parts: typing.Any = urlparse(REDIS_URL)
REDIS_DB = _redis_parts.path.strip('/')
REDIS_HOST, REDIS_PORT = _redis_parts.netloc.split(':')
REDIS_PORT: int = int(REDIS_PORT)

class PipelineType(enum.Enum):
  BOTTLE: str = 'Bottle'
  CONCURRENT: str = 'Concurrent'

logger = logging.getLogger(__name__)
logger.info(f'DEBUG[{DEBUG}]')
logger.info(f'RedisDB[{REDIS_DB}]')

DOCKER_SERVICE_NAME: str = 'bert-etl-redis'
DOCKER_REDIS_IMAGE: str = 'library/redis:latest'




#!/usr/bin/env python
'''
In this exmaple of how to use bert, we'll create a pipeline that'll download Messier Objects from a copy of the Messier
Catelog on wikipedia.
'''

# https://github.com/mdeff/fma#coverage
# http://www.audiocontentanalysis.org/data-sets/
# https://data.vision.ee.ethz.ch/cvl/ae_dataset/

import os
import typing

from bert import binding, constants, utils

ENCODING: str = 'utf-8'
OUTPUT_DIR: str = os.path.join(os.getcwd(), 'sounds')
HEADERS: typing.Dict[str, str] = {
  'User-Agent': 'https://github.com/jbcurtin/bert'
}

@binding.follow('noop', pipeline_type=constants.PipelineType.BOTTLE)
def sync_sounds():
  '''
  comm_binder takes the function you're currently writing as an argument and returns two queues backed by Redis.
  - The first Queue is a work queue, named after a the concept of accepting work
  - The second Queue is a done queue, named after the concept of offering work because you're done with it in this step
  - ologger is a python logger instance, with the associated process-id in the name that you're executing within
  '''
  work_queue, done_queue, ologger = utils.comm_binders(sync_sounds)

  '''
  The rationale put putting imports inside the fuctions is based aruond how the jobs are executed. When DEBUG is True,
  everything is synchronous. When DEBUG is False, the multiprocessing module is utilized to spawn daemons and manage
  subprocesses executing copies of the fuction you're working inside.
  '''
  import audioread
  import csv
  import librosa
  import os
  import requests
  import typing
  import zipfile
  '''
  Repeatability or idempotent functions go a long way with reducing your debug loop. In the commandline function of
  bert, you're provided with the option to flush the Redis database on every job start. This allows you to create
  artifacts while processing data. Binding the functions together, allowing you to execute the functions in order
  without having to recompute the resutls
  '''
  if not os.path.exists(OUTPUT_DIR):
    ologger.info(f'Creating Directory[{OUTPUT_DIR}]')
    os.makedirs(OUTPUT_DIR)

  '''
  Sometimes repeatability requires code-duplication. When opening an archive, it could be a few gigabytes in size. If the archive isn't
  open, the following code will open it to disk. If this function is ran a second time, it'll skip opening the files to disk and add the
  files to queue.
  ''' 
  archive_path: str = os.path.join(OUTPUT_DIR, 'ae_dataset.zip')
  if not os.path.exists(archive_path):
    # full dataset
    #  https://data.vision.ee.ethz.ch/cvl/ae_dataset/ae_dataset.zip
    url: str = f'https://github.com/jbcurtin/bert/blob/d4a0e0792f093fd86708074c2ff14ef0040aa1d1/data/ae_dataset-subset.zip?raw=true'
    ologger.info(f'Downloading Sounds[{url}]')
    response = requests.get(url, stream=True, headers=HEADERS)
    with open(archive_path, 'wb') as stream:
      for chunk in response.iter_content(chunk_size=1024):
        stream.write(chunk)
  
  archive_output_path: str = os.path.join(OUTPUT_DIR, 'ae_dataset')
  if not os.path.exists(archive_output_path):
    ologger.info(f'Opening Archive[{archive_output_path}]')
    with zipfile.ZipFile(archive_path, 'r') as compressed_stream:
      for name in compressed_stream.namelist():
        info: zipfile.ZipInfo = compressed_stream.getinfo(name)
        infopath: str = os.path.join(archive_output_path, info.filename.strip('/'))
        if info.is_dir():
          if not os.path.exists(infopath):
            os.makedirs(infopath)
          continue

        with open(infopath, 'wb') as output_stream:
          output_stream.write(compressed_stream.read(name))

        try:
          wave, samplerate = librosa.core.load(infopath, mono=True, sr=None)
        except audioread.NoBackendError:
          ologger.info(f'Unable to load WaveFile[{infopath}]')
          continue

        done_queue.put({
          'class': os.path.dirname(infopath).rsplit('/', 1)[1],
          'wave_path': infopath,
          'samplerate': samplerate,
          'duration': librosa.core.get_duration(wave),
        })

  else:
    ologger.info(f'Scanning Archive[{archive_output_path}]')
    with zipfile.ZipFile(archive_path, 'r') as compressed_stream:
      for name in compressed_stream.namelist():
        info: zipfile.ZipInfo = compressed_stream.getinfo(name)
        infopath: str = os.path.join(archive_output_path, info.filename.strip('/'))
        if info.is_dir():
          continue

        try:
          wave, samplerate = librosa.core.load(infopath, mono=True, sr=None)
        except audioread.NoBackendError:
          ologger.info(f'Unable to load WaveFile[{infopath}]')
          continue

        done_queue.put({
          'class': os.path.dirname(infopath).rsplit('/', 1)[1],
          'wave_path': infopath,
          'samplerate': samplerate,
          'duration': librosa.core.get_duration(wave),
        })


@binding.follow(sync_sounds, pipeline_type=constants.PipelineType.BOTTLE)
def analyse_sound_files():
  work_queue, done_queue, ologger = utils.comm_binders(analyse_sound_files)
  import librosa
  import numpy as np

  samples: typing.Dict[str, typing.List[np.array]] = []
  for details in work_queue:
    if details['class'] == 'test':
      continue

    # Begin work here
    ologger.info('New to Bert? Check out the tutorial[pending]')
    import sys; sys.exit(0)




#!/usr/env/bin python

import argparse
import docker
import functools
import importlib
import logging
import multiprocessing
import os
import signal
import time
import typing
import types


from bert import constants, utils

from datetime import datetime

from docker.errors import APIError

logger = logging.getLogger(__name__)

JOBS: typing.Dict[str, types.FunctionType] = {}
STOP_DAEMON: bool = False

def capture_options() -> typing.Any:
  parser = argparse.ArgumentParser()
  parser.add_argument('-n', '--new-module', default=None, required=False)
  parser.add_argument('-j', '--jobs', default=None, required=False)
  parser.add_argument('-d', '--debug', action='store_false', default=True)
  parser.add_argument('-m', '--module-name', default='bert')
  parser.add_argument('-o', '--how', default=False, action='store_true')
  parser.add_argument('-f', '--flush-db', action='store_true', default=False)
  parser.add_argument('-c', '--channel', default=0, type=int, help="Redis DB[0-16]; Default 0")
  parser.add_argument('-r', '--redis-port', default=6379, type=int)
  parser.add_argument('-l', '--log-error-only', default=True, action="store_false")
  parser.add_argument('-e', '--restart-job', default=True, action="store_false")
  parser.add_argument('-a', '--max-restart', default=10, type=int)
  return parser.parse_args()

def setup(options: argparse.Namespace) -> None:
  from bert import constants
  docker_client: typing.Any = docker.from_env()
  if not constants.DOCKER_SERVICE_NAME in [c.name for c in docker_client.containers.list()]:
    logger.info('Starting RedisClient')
    try:
      docker_client.containers.run('library/redis:latest', name=constants.DOCKER_SERVICE_NAME, detach=True, ports={
      6379: options.redis_port
    })
    except APIError as err:
      logger.info("Redis is already running under another service-name. Let's use that.")

  if options.flush_db:
    import redis
    redis_client = redis.Redis(host=constants.REDIS_HOST, port=constants.REDIS_PORT, db=options.channel)
    logger.info(f'Flushing Redis DB[{options.channel}]')
    redis_client.flushdb()

def scan_jobs(options):
  global JOBS
  module = importlib.import_module(f'{options.module_name}.jobs')
  for member_name in dir(module):
    if member_name.startswith('_'):
      continue

    member = getattr(module, member_name)
    if type(member) != types.FunctionType:
      continue

    JOBS[member_name] = member

def validate_jobs(options) -> None:
  for job in options.jobs.split(','):
    if job not in JOBS.keys():
      raise NotImplementedError(f'Job[{job}] not found. Available Jobs[{", ".join(JOBS.keys())}]')

def handle_signal(sig, frame):
  if sig == 2:
    global STOP_DAEMON
    STOP_DAEMON = True
    import sys; sys.exit(0)

  else:
    logger.info('Unhandled Signal[{sig}]')

def start_jobs(options):
  signal.signal(signal.SIGINT, handle_signal)
  from bert import binding
  if constants.DEBUG:
    job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
    for job in job_chain:
      logger.info(f'Running Job[{job.func_space}] as [{job.pipeline_type.value}] for [{job.__name__}]')
      job()

  else:
    job_chain: typing.List[types.FunctionType] = binding.build_job_chain()
    for job in job_chain:
      logger.info(f'Running Job[{job.func_space}] as [{job.pipeline_type.value}] for [{job.__name__}]')
      logger.info(f'Job worker count[{job.workers}]')
      processes: typing.List[multiprocessing.Process] = []

      @functools.wraps(job)
      def _job_runner(*args, **kwargs):
        job_restart_count: int = 0
        while job_restart_count < options.max_restart:
          try:
            job(*args, **kwargs)
          except Exception as err:
            if options.log_error_only:
              logger.exception(err)

            else:
              raise err

          job_restart_count += 1
        else:
          logger.exception(f'Job[{job}] failed {job_restart_count} times')

      for idx in range(0, job.workers):
        logging.info(f'Spawning Process[{idx}]')
        proc: multiprocessing.Process = multiprocessing.Process(target=_job_runner, args=())
        proc.daemon = True
        proc.start()
        processes.append(proc)

      else:
        while not STOP_DAEMON and any([proc.is_alive() for proc in processes]):
          time.sleep(constants.DELAY)

if __name__ in ['__main__']:
  options = capture_options()
  setup(options)
  scan_jobs(options)
  validate_jobs(options)
  start_jobs(options)




#!/usr/env/bin python

import logging
import docker
import hashlib
import json
import multiprocessing
import os
import redis
import shutil
import subprocess
import time
import typing
import types

import requests

from datetime import datetime

from bert import constants

from urllib.parse import urlparse, ParseResult

logger = logging.getLogger(__name__)

class QueuePacker(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, datetime):
      return obj.strftime(constants.DATETIME_FORMAT)

    return super(QueuePacker, self).default(obj)

class Queue:
  DEFAULT_DELAY: int = 1728000
  @staticmethod
  def Pack(datum: typing.Dict[str, typing.Any]) -> str:
    return json.dumps(datum, cls=QueuePacker)

  @staticmethod
  def UnPack(datum: str) -> typing.Dict[str, typing.Any]:
    return json.loads(datum)

  def __init__(self, redis_key, redis_url: str=constants.REDIS_URL):
    self._key = redis_key
    self._redis_client = None

    parts: ParseResult = urlparse(redis_url)
    self._db = parts.path.strip('/')
    self._host, self._port = parts.netloc.split(':')
    self._port = int(self._port)
    self._redis_client = redis.Redis(host=self._host, port=self._port, db=self._db)

  def __iter__(self):
    return self

  def __next__(self):
    value = self.get()
    if value is None or value == 'STOP':
      raise StopIteration

    return value

  def peg(self, key: str, delay: str=DEFAULT_DELAY) -> bool:
    value: str = self._redis_client.get(key)
    if value is None:
      self._redis_client.set(key, 1)
      self._redis_client.expire(key, delay)
      return True

    return False

  def flushdb(self) -> None:
    self._redis_client.flushdb()

  def put_block(self, value: str, delay: int=1728000) -> None:
    value_sorted = ''.join(sorted(value))
    value_sorted = ''.join(['put-block', value_sorted])
    value_hash = hashlib.md5(value_sorted.encode('utf-8')).hexdigest()
    if self._redis_client.get(value_hash) is None:
      self._redis_client.set(value_hash, 1)
      self._redis_client.expire(value_hash, delay)
      self.put(value)

  def size(self) -> int:
    return int(self._redis_client.llen(self._key))

  def get(self) -> typing.Dict[str, typing.Any]:
    try:
      value: str = self._redis_client.lpop(self._key).decode('utf-8')
    except AttributeError:
      return 'STOP'

    else:
      return self.__class__.UnPack(value)

  def put(self, value: typing.Dict[str, typing.Any]) -> None:
    value: str = self.__class__.Pack(value)

    self._redis_client.rpush(self._key, value.encode('utf-8'))

  def bulk(self, stop_iter_key: str, bulk_count: int=50000) -> typing.List[str]:

    index: int = 0
    datums: typing.List[str] = []
    while index < bulk_count:
      datum: str = self._redis_client.lpop(self._key)
      if datum is None:
        break

      if datum == stop_iter_key:
        break

      datums.append(datum.decode('utf-8'))

    logging.info(len(datums))
    return datums

def find_in_elem(elem, *args) -> str:
  base = elem
  for key in args:
    try:
      base = base.find(fix_tag(key))
    except Exception as err:
      return 0

  else:
    if base == None:
      return 0

    elif base.text == None:
      return 0

    elif isinstance(base.text, bytes):
      return base.text.decode('utf-8')

    elif isinstance(base.text, str):
      return base.text

    else:
      import ipdb; ipdb.set_trace()
      return 0

def parse_timestamp(input_datetime: str) -> datetime:
  if input_datetime == 0:
    return None

  return datetime.strptime(input_datetime, '%Y-%m-%dT%H:%M:%SZ')

def fix_schema_tag(tag) -> str:
  return '{http://www.w3.org/2001/XMLSchema}%s' % tag

def fix_tag(tag) -> str:
  return '{http://www.mediawiki.org/xml/export-0.10/}%s' % tag

def zero_file(filepath, file_size):
  with open(filepath, 'wb') as stream:
    for chunk in iter(lambda: b'\0' * 4096, b''):
      stream.write(chunk)
      if stream.tell() > file_size:
        break

def run_command(cmd: str, allow_error: typing.List[int] = [0]) -> str:
  ologger = logging.getLogger('.'.join([__name__, multiprocessing.current_process().name]))
  cmd: typing.List[str] = cmd.split(' ')
  proc = subprocess.Popen(' '.join(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
  while proc.poll() is None:
    time.sleep(.1)

  if proc.poll() > 0:
    if not proc.poll() in allow_error:
      raise NotImplementedError(f'{proc.poll()}, {proc.stderr.read()}')

    return proc.stderr.read().decode(constants.ENCODING)

  return proc.stdout.read().decode('utf-8')

def build_dockerfile(dockerfile_dir: str, build_filename: str, build_name: str, version: str='latest') -> typing.Tuple[docker.DockerClient, str]:
  client = docker.DockerClient()
  dockerfile_path = os.path.join(dockerfile_dir, build_filename)
  if not os.path.exists(dockerfile_path):
    raise IOError(f'Missing Dockerfile[{dockerfile_path}]')

  docker_tags: typing.List[str] = [item for sublist in [image.tags for image in client.images.list()] for item in sublist]
  docker_tag = f'{build_name}:{version}'
  if not docker_tag in docker_tags \
      or constants.DOCKER_NO_CACHE:
    ologger = logging.getLogger('.'.join([__name__,'build_dockerfile', multiprocessing.current_process().name]))
    ologger.info(f'Building Docker Image[{docker_tag}]')
    client.images.build(
        nocache=constants.DOCKER_NO_CACHE,
        path=dockerfile_dir,
        dockerfile=dockerfile_path,
        tag=docker_tag)

def comm_binders(func: types.FunctionType) -> typing.Tuple[Queue, Queue, 'ologger']:
    ologger = logging.getLogger('.'.join([func.__name__, multiprocessing.current_process().name]))
    return Queue(func.work_key), Queue(func.done_key), ologger

def new_module(options: typing.Any) -> None:
  import bert
  operating_dir: str = os.path.join(os.getcwd(), options.new_module)
  if not os.path.exists(operating_dir):
    os.makedirs(operating_dir)

  bert_jobs_filepath: str = os.path.join(os.path.dirname(bert.__file__), 'jobs.py')
  jobs_file: str = os.path.join(operating_dir, 'jobs.py')
  if not os.path.exists(jobs_file):
    shutil.copyfile(bert_jobs_filepath, jobs_file)

  logger.info(f'''
  Run the new module with the following command:
  PYTHONPATH='.' bert-runner.py -j sync_sounds -m {options.new_module}
  ''')

def show_how(options: typing.Any) -> None:
  logger.info('''
  Run the new module with the following command:
  PYTHONPATH='.' bert-runner.py -j <job_name> -m <module_name>
  ''')




#!/usr/bin/env python

import sys

from bert import factory, utils

if __name__ in ['__main__']:
  options = factory.capture_options()
  if options.how:
    utils.show_how(options)
    sys.exit(0)

  if options.new_module:
    utils.new_module(options)
    sys.exit(0)

  factory.setup(options)
  factory.scan_jobs(options)
  factory.validate_jobs(options)
  factory.start_jobs(options)

